if not SetCameraFov or GetInt( "savegame.mod.legacy" ) == 1 then

	function init()
		FlyActive = false
		Target = nil
		NoclipActive = false
		NoclipTransform = nil
		NoclipAngle = nil
	end

	local function toEuler( quat )
		local x, y, z, w = quat[1], quat[2], quat[3], quat[4]
		-- Credit to https://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm

		local bank, heading, attitude

		local s = 2 * x * y + 2 * z * w
		if s >= 1 then
			heading = 2 * math.atan2( x, w )
			bank = 0
			attitude = math.pi / 2
		elseif s <= -1 then
			heading = -2 * math.atan2( x, w )
			bank = 0
			attitude = math.pi / -2
		else
			bank = math.atan2( 2 * x * w - 2 * y * z, 1 - 2 * x ^ 2 - 2 * z ^ 2 )
			heading = math.atan2( 2 * y * w - 2 * x * z, 1 - 2 * y ^ 2 - 2 * z ^ 2 )
			attitude = math.asin( s )
		end

		return math.deg( bank ), math.deg( heading ), math.deg( attitude )
	end

	local function aorb( a, b, d )
		return (a and d or 0) - (b and d or 0)
	end

	local function getDelta( dt, tr )
		local f, b, l, r, s = InputDown( "up" ), InputDown( "down" ), InputDown( "left" ), InputDown( "right" ),
		                      InputDown( "space" )
		if f or b or l or r or s then
			local dist = InputDown( "ctrl" ) and 0.01 or InputDown( "shift" ) and 0.3 or 0.1
			local fup = s and dist or 0
			return VecAdd( Vec( 0, (f and 0 or fup) + dt / 10, 0 ),
			               TransformToParentVec( tr, Vec( aorb( r, l, dist ), (f and fup or 0), aorb( b, f, dist ) ) ) )
		end
	end

	function tick( dt )
		if InputPressed( "v" ) then
			FlyActive = not FlyActive
			Target = nil
			NoclipActive = false
		end
		if not FlyActive then
			return
		end

		local height = InputDown( "ctrl" ) and 0.85 or 1.7
		if NoclipActive then
			if not Target then
				Target = NoclipTransform.pos
			end
			local dx, dy = InputValue( "mousedx" ), InputValue( "mousedy" )
			local sensitivity = GetInt( "options.input.sensitivity" ) / 100
			NoclipAngle[1] = math.min( math.max( NoclipAngle[1] - dy / 10 * sensitivity, -80 ), 80 )
			NoclipAngle[2] = NoclipAngle[2] - dx / 10 * sensitivity
			NoclipTransform.rot = QuatEuler( NoclipAngle[1], NoclipAngle[2], 0 )
			local d = getDelta( 0, NoclipTransform )
			if d then
				Target = VecAdd( Target, d )
				NoclipTransform.pos = Target
			end
			SetCameraTransform( Transform( VecAdd( NoclipTransform.pos, { 0, height, 0 } ), NoclipTransform.rot ) )
			SetPlayerTransform( NoclipTransform )
		else
			---[[
			local current_pos = GetPlayerTransform().pos
			local v = GetPlayerVelocity()
			if not Target or VecLength( v ) > 1 then
				Target = current_pos
			end

			local d = getDelta( dt, GetCameraTransform() )
			if d then
				Target = VecAdd( Target, d )
			end
			SetPlayerVelocity( VecScale( VecSub( Target, current_pos ), 100 ) )
			-- ]]
			--[[
			local current_pos = GetPlayerTransform().pos
			if not Target then Target = current_pos end
			local d = getDelta(0, GetCameraTransform())
			if d then Target = VecAdd(Target, VecScale(d, 60 * dt)) end
			SetPlayerVelocity(VecScale(VecSub(Target, current_pos), 2 / dt))
			--]]
		end

		local nocliphit = GetInt( "savegame.mod.hybrid" ) == 0 and
						                  (QueryRaycast( VecAdd( Target, { 0, 0.3, 0 } ), { 0, 1, 0 }, height - .3, 0.2 ) or
										                  QueryRaycast( VecAdd( Target, { 0, height, 0 } ), { 0, -1, 0 }, height - .3, 0.2 ))
		if nocliphit ~= NoclipActive then
			if NoclipActive then
				NoclipTransform = nil
				NoclipAngle = nil
			else
				NoclipTransform = GetPlayerTransform()
				NoclipAngle = Vec( toEuler( GetCameraTransform().rot ) )
			end
			NoclipActive = not NoclipActive
		end
	end

else

	function init()
		FlyActive = false

		MinFrameTime = GetInt( "savegame.mod.noslowmo" ) == 1 and 1 / 60 or 0
	end

	local function aorb( a, b, d )
		return (a and d or 0) - (b and d or 0)
	end

	local function getDelta( tr )
		local f, b, l, r, s = InputDown( "up" ), InputDown( "down" ), InputDown( "left" ), InputDown( "right" ),
		                      InputDown( "space" )
		if f or b or l or r or s then
			local dist = InputDown( "ctrl" ) and 0.02 or InputDown( "shift" ) and 0.5 or 0.2
			local fup = s and dist or 0
			return VecAdd( Vec( 0, f and 0 or fup, 0 ),
			               TransformToParentVec( tr, Vec( aorb( r, l, dist ), (f and fup or 0), aorb( b, f, dist ) ) ) )
		end
	end

	function tick( dt )
		dt = math.max( dt, MinFrameTime )
		if InputPressed( "v" ) then
			FlyActive = not FlyActive
			FlyTransform = nil
			if not FlyActive and LastV then
				SetPlayerVelocity( VecScale( LastV, 60 ) )
			end
		end
		if not FlyActive then
			return
		end
		local ct = GetPlayerTransform( true )
		if not FlyTransform then
			FlyTransform = ct
		end
		FlyTransform.rot = ct.rot
		local d = getDelta( FlyTransform )
		if d then
			FlyTransform.pos = VecAdd( FlyTransform.pos, VecScale( d, 60 * dt ) )
		end
		if VecLength( VecSub( FlyTransform.pos, ct.pos ) ) > 2 then
			FlyTransform = ct
		end
		SetPlayerTransform( FlyTransform, true )
		LastV = d
	end

end
